@@ -5,7 +5,7 @@
 
 
 //default font.. -font at runtime should work
-#define USE_FONT        "fixed"
+#define USE_FONT        "*"
 
 //default positions.. can be changed with -g at runtime
 #define STD_WIDTH       730
@@ -13,5 +13,5 @@
 #define LOC_X           30
 #define LOC_Y           30
 
-#define VERSION "1.1"
+#define VERSION "1.2"
 
