@@ -68,6 +68,7 @@ struct logfile_entry
   ino_t inode;                  /* inode of the file opened                     */
   off_t last_size;              /* file size at the last check                  */
   unsigned long color;          /* color to be used for printing                */
+  const char *colorname;        /* color name/string                            */
   int partial;                  /* true if the last line isn't complete         */
   int lastpartial;              /* true if the previous output wasn't complete  */
   struct line_node *last;       /* last line we output                          */
@@ -124,7 +125,7 @@ struct timeval interval = { 2, 400000 };
 /* command line options */
 int opt_noinitial, opt_shade, opt_frame, opt_reverse, opt_nofilename,
   opt_outline, opt_noflicker, opt_whole, opt_update, opt_wordwrap,
-  opt_justify, geom_mask, reload = 0;
+  opt_justify, geom_mask, opt_minspace, reload;
 const char *command = NULL,
   *fontname = USE_FONT, *dispname = NULL, *def_color = DEF_COLOR,
   *continuation = "|| ", *cont_color = DEF_CONT_COLOR;
@@ -153,6 +154,9 @@ void list_files (int);
 void force_reopen (int);
 void force_refresh (int);
 void blank_window (int);
+#ifdef USE_TOON_GET_ROOT_WINDOW
+Window ToonGetRootWindow(Display *, int, Window *);
+#endif /* USE_TOON_GET_ROOT_WINDOW */
 
 void InitWindow (void);
 unsigned long GetColor (const char *);
@@ -223,68 +227,60 @@ GetColor (const char *ColorName)
   return Color.pixel;
 }
 
-static Window
-root_window (Display * display, int screen_number)
+#ifndef USE_TOON_GET_ROOT_WINDOW
+static void
+find_root_window (Display *display, int screen_number)
 {
-  Atom SWM_VROOT = XInternAtom (display, "__SWM_VROOT", False);
-  Window real_root_window = RootWindow (display, screen_number);
-
-  if (root)                     /* root window set via option */
-    return root;
-
-  if (SWM_VROOT != None)
+  if (!root)
     {
-      Window unused, *windows;
+      Atom SWM_VROOT = XInternAtom (display, "__SWM_VROOT", False);
+      Atom NAUTILUS_DESKTOP_WINDOW_ID = XInternAtom (display, "NAUTILUS_DESKTOP_WINDOW_ID", False);
+      root = RootWindow (display, screen_number);
+
+      Window unused, *windows = 0;
       unsigned int count;
 
-      if (XQueryTree (display, real_root_window, &unused, &unused, &windows, &count))
+      Atom type;
+      int format;
+      unsigned long nitems, bytes_after_return;
+      unsigned char *virtual_root_window;
+
+      if (XGetWindowProperty (display, root, NAUTILUS_DESKTOP_WINDOW_ID,
+                              0, 1, False, XA_WINDOW, &type, &format,
+                              &nitems, &bytes_after_return,
+                              &virtual_root_window) == Success
+          && type == XA_WINDOW)
+        {
+          if (XQueryTree (display, *(Window *)virtual_root_window, &unused, &unused, &windows, &count))
+            root = windows[count - 1];
+
+          XFree (virtual_root_window);
+        }
+      else if (XQueryTree (display, root, &unused, &unused, &windows, &count))
         {
           int i;
 
           for (i = 0; i < count; i++)
             {
-              Atom type;
-              int format;
-              unsigned long nitems, bytes_after_return;
-              unsigned char *virtual_root_window;
-
               if (XGetWindowProperty (display, windows[i], SWM_VROOT,
                                       0, 1, False, XA_WINDOW, &type, &format,
                                       &nitems, &bytes_after_return,
-                                      &virtual_root_window) == Success)
+                                      &virtual_root_window) == Success
+                   && type == XA_WINDOW)
                 {
-                  if (type != None)
-                    {
-                      if (type == XA_WINDOW)
-                        {
-                          root = *(Window *)virtual_root_window;
-                          XFree (windows);
-                          XFree (virtual_root_window);
-                          return root;
-                        }
-                      else
-                        fprintf (stderr, "__SWM_VROOT property type mismatch");
-                    }
+                  root = *(Window *)virtual_root_window;
+                  XFree (virtual_root_window);
+                  break;
                 }
-              else
-                fprintf (stderr,
-                         "failed to get __SWM_VROOT property on window 0x%lx",
-                         windows[i]);
             }
 
-          if (count)
-            XFree (windows);
+          XFree (windows);
         }
       else
-        fprintf (stderr, "Can't query tree on root window 0x%lx",
-                 real_root_window);
+        fprintf (stderr, "Can't query tree on root window 0x%lx", root);
     }
-  else
-    /* This shouldn't happen. The Xlib documentation is wrong BTW. */
-    fprintf (stderr, "Can't intern atom __SWM_VROOT");
-
-  return real_root_window;
 }
+#endif /* USE_TOON_GET_ROOT_WINDOW */
 
 void
 InitWindow (void)
@@ -304,12 +300,13 @@ InitWindow (void)
   ScreenHeight = DisplayHeight (disp, screen);
   ScreenWidth = DisplayWidth (disp, screen);
 
-  root = root_window (disp, screen);
+  find_root_window (disp, screen);
 
   gcm = GCBackground;
   gcv.graphics_exposures = True;
   WinGC = XCreateGC (disp, root, gcm, &gcv);
   XMapWindow (disp, root);
+
   XSetForeground (disp, WinGC, GetColor (DEF_COLOR));
 
   for (e = loglist; e; e = e->next)
@@ -342,6 +339,19 @@ InitWindow (void)
         e->font_height = xfe->max_logical_extent.height;
         e->font_ascent = -xfe->max_logical_extent.y;
       }
+
+      if (e->font_height > height - effect_y_space)
+	{
+	  fprintf(stderr, "\n  the display isn't tall enough to display a single line in font '%s'\n",
+		  e->fontname);
+	  fprintf(stderr, "\n  the geometry in use is %d pixels tall\n", height);
+	  fprintf(stderr, "\n  font '%s' is %d pixels tall\n", e->fontname, e->font_height);
+	  if (effect_y_space)
+	    fprintf(stderr, "\n  the shade or outline options need an extra %d pixel%s of vertical space\n",
+		    effect_y_space, effect_y_space == 1 ? "" : "s");
+	  fprintf(stderr, "\n");
+	  exit(1);
+	}
     }
 
   if (geom_mask & XNegative)
@@ -349,31 +359,12 @@ InitWindow (void)
   if (geom_mask & YNegative)
     win_y = win_y + ScreenHeight - height;
 
-  if (opt_outline)
-    {
-      /* adding outline increases the total width and height by 2
-         pixels each, and offsets the text one pixel right and one
-         pixel down */
-      effect_x_space = effect_y_space = 2;
-      effect_x_offset = effect_y_offset = 1;
-    }
-  else if (opt_shade)
-    {
-      /* adding a shadow increases the space used */
-      effect_x_space = abs (SHADE_X);
-      effect_y_space = abs (SHADE_Y);
+  {
+    struct logfile_entry *e;
 
-      /* if the shadow is to the right and below then we don't need
-       * to move the text to make space for it, but shadows to the left
-       * and above need accomodating */
-      effect_x_offset = SHADE_X > 0 ? 0 : -SHADE_X;
-      effect_y_offset = SHADE_Y > 0 ? 0 : -SHADE_Y;
-    }
-  else
-    {
-      effect_x_space = effect_y_space = 0;
-      effect_x_offset = effect_y_offset = 0;
-    }
+    for (e = loglist; e; e = e->next)
+      e->color = GetColor (e->colorname);
+  }
 
   XSelectInput (disp, root, ExposureMask | FocusChangeMask);
 }
@@ -571,6 +562,12 @@ refresh (int miny, int maxy, int clear, int refresh_all)
 #endif
     }
 
+  /* at least one of the lines must fit in the allocated area.  we've
+   * already checked at initialisation time that all the fonts are small
+   * enough to fit at least one line in the display area, but assert it
+   * again here to be sure */
+  assert(line != linelist);
+
   /* any lines that didn't just get looked at are never going to be, so break the chain */
   if (line) line->prev->next = 0;
 
@@ -655,12 +652,12 @@ transform_line (char *s)
 char *
 concat_line (char *p1, const char *p2)
 {
-  assert(p2);
-
   int l1 = p1 ? strlen (p1) : 0;
   int l2 = strlen (p2);
   char *r;
 
+  assert (p2);
+
   if (p1)
     r = xrealloc(p1, l1 + l2 + 1);
   else
@@ -925,6 +922,8 @@ possibly_split_long_line (struct logfile_entry *log)
       int wrapped = 0;
       char *break_p = NULL;
       int width_at_break_p = 0;
+      int prefix_len;
+
       spaces = 0;
 
       if (opt_justify)
@@ -992,8 +991,6 @@ possibly_split_long_line (struct logfile_entry *log)
       if (!wrapped)
         break;
 
-      int prefix_len;
-
       /* choose where to break the line */
       if (opt_wordwrap && break_p && break_p != beg)
         {
@@ -1161,13 +1158,16 @@ main_loop (void)
 
               /* print filename if any, and if last line was from
                * different file */
-              if (!opt_nofilename && lastprinted != current && current->desc[0])
-                {
+              if (lastprinted != current)
+		{
                   current->last = 0;
-                  insert_new_line (xstrdup ("["), current);
-                  append_to_existing_line (xstrdup (current->desc), current);
-                  append_to_existing_line (xstrdup ("]"), current);
-                }
+		  if (!opt_nofilename && current->desc[0])
+		    {
+		      insert_new_line (xstrdup ("["), current);
+		      append_to_existing_line (xstrdup (current->desc), current);
+		      append_to_existing_line (xstrdup ("]"), current);
+		    }
+		}
 
               /* if we're dealing with partial lines, and the last
                * time we showed the line it wasn't finished ... */
@@ -1285,11 +1285,6 @@ main (int argc, char *argv[])
 
   setlocale (LC_CTYPE, "");     /* try to initialize the locale. */
 
-  /* window needs to be initialized before colorlookups can be done */
-  /* just a dummy to get the color lookups right */
-  geom_mask = NoValue;
-  InitWindow ();
-
   for (i = 1; i < argc; i++)
     {
       const char *arg = argv[i];
@@ -1334,6 +1329,8 @@ main (int argc, char *argv[])
             opt_shade = 1;
           else if (!strcmp (arg, "-outline"))
             opt_outline = 1;
+          else if (!strcmp (arg, "-minspace"))
+            opt_minspace = 1;
           else if (!strcmp (arg, "-noflicker"))
             opt_noflicker = 1;
           else if (!strcmp (arg, "-frame"))
@@ -1357,7 +1354,12 @@ main (int argc, char *argv[])
           else if (!strcmp (arg, "-noinitial"))
             opt_noinitial = 1;
           else if (!strcmp (arg, "-id"))
-            root = atoi (argv[++i]);
+            {
+              unsigned long id;
+
+              if (sscanf (argv[++i], "%li", &id) == 1 && id)
+                root = id;
+            }
           else if (!strcmp (arg, "-interval") || !strcmp (arg, "-i"))
             {
               double iv = atof (argv[++i]);
@@ -1422,7 +1424,7 @@ main (int argc, char *argv[])
               e->desc = xstrdup (desc);
             }
 
-          e->color = GetColor (fcolor);
+          e->colorname = fcolor;
           e->partial = 0;
           e->fontname = fontname;
           e->last = NULL;
@@ -1493,6 +1495,32 @@ main (int argc, char *argv[])
     }
 #endif
 
+  if (opt_outline && !opt_minspace)
+    {
+      /* adding outline increases the total width and height by 2
+         pixels each, and offsets the text one pixel right and one
+         pixel down */
+      effect_x_space = effect_y_space = 2;
+      effect_x_offset = effect_y_offset = 1;
+    }
+  else if (opt_shade && !opt_minspace)
+    {
+      /* adding a shadow increases the space used */
+      effect_x_space = abs (SHADE_X);
+      effect_y_space = abs (SHADE_Y);
+
+      /* if the shadow is to the right and below then we don't need
+       * to move the text to make space for it, but shadows to the left
+       * and above need accomodating */
+      effect_x_offset = SHADE_X > 0 ? 0 : -SHADE_X;
+      effect_y_offset = SHADE_Y > 0 ? 0 : -SHADE_Y;
+    }
+  else
+    {
+      effect_x_space = effect_y_space = 0;
+      effect_x_offset = effect_y_offset = 0;
+    }
+
   InitWindow ();
 
   install_signal (SIGINT, blank_window);
@@ -1581,9 +1609,11 @@ display_help (char *myname)
           " -partial                  show lines even if they don't end with a \\n\n"
           " -update                   allow updates to old partial lines\n"
           " -cont                     string to prefix continued partial lines with\n"
+          "                           defaults to \"|| \"\n"
           " -wordwrap                 wrap long lines at spaces to avoid breaking words\n"
-          "                           defaults to \"[+]\"\n"
           " -shade                    add shading to font\n"
+          " -outline                  add black outline to font\n"
+          " -minspace                 force minimum line spacing\n"
           " -noinitial                don't display the last file lines on\n"
           "                           startup\n"
           " -i | -interval seconds    interval between checks (fractional\n"
@@ -1623,5 +1653,3 @@ daemonize (void)
 
   return 0;
 }
-
-/* todo - get reverse display working again */
